#pragma once

#include "nodes/applied_bonus_points.hpp"
#include "nodes/bonus_points.hpp"
#include "nodes/changed_max_points.hpp"
#include "nodes/combined_grade_average.hpp"
#include "nodes/combined_grade_list.hpp"
#include "nodes/dynexite_exam.hpp"
#include "nodes/exam_stats.hpp"
#include "nodes/first_try.hpp"
#include "nodes/grade_list.hpp"
#include "nodes/node.hpp"
#include "nodes/passing_limit.hpp"
#include "nodes/student_data.hpp"
#include "nodes/student_points.hpp"
#include "nodes/user_grade_boundaries.hpp"
#include "nodes/user_grade_config.hpp"
#include "nodes/user_point_config.hpp"

#include "util/enums.hpp"

#include <cassert>
#include <tuple>

template <typename... ChangedNode>
void updateNodesIfNecessary() noexcept {}

template <typename... ChangedNodes, typename N, typename... Nodes>
void updateNodesIfNecessary(N &n,
                            Nodes &...nodes) noexcept { /* TODO: noexcept? */
    if constexpr (AnyIsPredecessorOfAny<std::tuple<ChangedNodes...>,
                                        std::tuple<N>>::value) {
        n.recalculate();
    }
    updateNodesIfNecessary<ChangedNodes...>(nodes...);
}

struct DAG {

    UserPointConfig mUserPointConfig{};
    UserGradeConfig mUserGradeConfig{};
    DynexiteExam mExam{};
    StudentData mStudents{};
    BonusPoints mBonus{};

    ExamStats mStats{mExam, mStudents};

    ChangedMaxPoints<GradingType::Closed> mChangedMaxPointsClosed{
        mUserPointConfig, mExam};
    ChangedMaxPoints<GradingType::Open> mChangedMaxPointsOpen{mUserPointConfig,
                                                              mExam};

    AppliedBonusPoints<GradingType::Closed> mAppliedBonusPointsClosed{
        mBonus, mChangedMaxPointsClosed, mChangedMaxPointsOpen};
    AppliedBonusPoints<GradingType::Open> mAppliedBonusPointsOpen{
        mBonus, mChangedMaxPointsClosed, mChangedMaxPointsOpen};

    StudentPoints<GradingType::Closed> mStudentPointsClosed{
        mExam, mAppliedBonusPointsClosed};
    StudentPoints<GradingType::Open> mStudentPointsOpen{
        mExam, mAppliedBonusPointsOpen};
    CombinedStudentPoints mStudentPointsCombined{mStudentPointsClosed,
                                                 mStudentPointsOpen};

    FirstTry mFirstTry{mExam, mStudents};

    PassingLimit<GradingType::Closed> mClosedPassingLimit{
        mChangedMaxPointsClosed, mFirstTry, mUserPointConfig};
    PassingLimit<GradingType::Open> mOpenPassingLimit{mChangedMaxPointsOpen,
                                                      mUserPointConfig};

    UserGradeBoundaries<GradingType::Closed> mClosedUserBoundaries{
        mUserPointConfig, mChangedMaxPointsClosed, mClosedPassingLimit};
    UserGradeBoundaries<GradingType::Open> mOpenUserBoundaries{
        mUserPointConfig, mChangedMaxPointsOpen, mOpenPassingLimit};

    GradeList<GradingType::Closed> mClosedGrades{
        mUserPointConfig, mUserGradeConfig, mClosedUserBoundaries, mExam,
        mStudentPointsClosed};
    GradeList<GradingType::Open> mOpenGrades{mUserPointConfig, mUserGradeConfig,
                                             mOpenUserBoundaries, mExam,
                                             mStudentPointsOpen};

    CombinedGradeList mGrades{mChangedMaxPointsClosed, mChangedMaxPointsOpen,
                              mClosedGrades, mOpenGrades, mUserGradeConfig};

    CombinedGradeAverage mGradeAverage{mGrades};

    template <typename... ChangedNodes>
    void updateMemberNodesIfNecessary() noexcept /* TODO: noexcept? */
    {
        updateNodesIfNecessary<ChangedNodes...>(
            mUserPointConfig, mUserGradeConfig, mExam, mStudents, mBonus,
            mStats, mChangedMaxPointsClosed, mChangedMaxPointsOpen,
            mAppliedBonusPointsClosed, mAppliedBonusPointsOpen,
            mStudentPointsClosed, mStudentPointsOpen, mStudentPointsCombined,
            mFirstTry, mClosedPassingLimit, mOpenPassingLimit,
            mClosedUserBoundaries, mOpenUserBoundaries, mClosedGrades,
            mOpenGrades, mGrades, mGradeAverage);
    }

    DAG(DynexiteExam exam, StudentData students, BonusPoints bonus,
        UserGradeConfig gradeConfig, UserPointConfig pointConfig)
        : mUserPointConfig(std::move(pointConfig)),
          mUserGradeConfig(std::move(gradeConfig)), mExam(std::move(exam)),
          mStudents(std::move(students)), mBonus(std::move(bonus)) {
        updateMemberNodesIfNecessary<DynexiteExam, StudentData, BonusPoints,
                                     UserGradeConfig, UserPointConfig>();
    }

    // copy constructor: bind reference to this object, not to the copied-from
    // object
    DAG(const DAG &other)
        : mUserPointConfig(other.mUserPointConfig),
          mUserGradeConfig(other.mUserGradeConfig), mExam(other.mExam),
          mStudents(other.mStudents), mBonus(other.mBonus) {
        updateMemberNodesIfNecessary<DynexiteExam, StudentData, BonusPoints,
                                     UserGradeConfig, UserPointConfig>();
    }
    // todo: move constructor etc

    // helper functions for easier access
    [[nodiscard]] constexpr const GradeBoundary &
    getPassingLimit(const GradingType grading) const noexcept {
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case GradingType::Closed:
            return mClosedPassingLimit.mPassingLimit;
        case GradingType::Open:
            return mOpenPassingLimit.mPassingLimit;
        }
    }
    [[nodiscard]] constexpr const EnumStorage<GradeBoundary, Grade> &
    getUserGradeBoundaries(const GradingType grading) const noexcept {
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case GradingType::Closed:
            return mClosedUserBoundaries.mBoundaries;
        case GradingType::Open:
            return mOpenUserBoundaries.mBoundaries;
        }
    }
    [[nodiscard]] constexpr const QList<Grade> &
    getGradeList(const GradingType grading) const noexcept {
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case GradingType::Closed:
            return mClosedGrades.mGrades;
        case GradingType::Open:
            return mOpenGrades.mGrades;
        case GradingType::Combined:
            return mGrades.mGrades;
        }
    }
    [[nodiscard]] constexpr auto &
    getPointsWithBonus(const GradingType grading) const noexcept {
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case GradingType::Closed:
            return mStudentPointsClosed.mPointsWithBonus;
        case GradingType::Open:
            return mStudentPointsOpen.mPointsWithBonus;
        case GradingType::Combined:
            return mStudentPointsCombined.mPointsWithBonus;
        }
    }

    /**
     * \brief Whether the given GradingType is present in this exam
     */
    [[nodiscard]] constexpr auto
    hasType(const GradingType grading) const noexcept {
        using enum GradingType;
        switch (grading) {
        default:
            assert(false);
            [[fallthrough]];
        case Closed:
            return mExam.mPointsMax[Closed] > Points{0.};
        case Open:
            return mExam.mPointsMax[Open] > Points{0.};
        case Combined:
            return hasType(Closed) && hasType(Open);
        }
    }

    [[nodiscard]] auto hasBonus() const noexcept {
        return !mBonus.mTotalBonusPointsPerStudent.empty();
    }
};
