/**
 * \file defaults.hpp
 * \brief Constant values used throughout the app.
 */

#pragma once

#include <QColor>

#include "grade_boundary_step.hpp"
#include "storage_helper.hpp"

namespace defaults {

/** \brief Whether to show an option to change grade boundary rounding. */
inline constexpr bool showRoundingSetting = false;

/** \brief Default grade boundary rounding  */
inline constexpr GradeBoundaryRounding boundaryRounding =
    GradeBoundaryRounding::Down;

/** \brief Whether to show an option to require passing the closed and open part
 * to pass the exam.
 *
 * The updated grading guidelines (march 2023) deem this not permissible.
 */
inline constexpr bool showRequireBothPartsSetting = false;

/** \brief Require passing both open and closed part to pass the exam.
 *
 * The updated grading guidelines (march 2023) deem this not permissible
 */
inline constexpr bool requirePassAllParts = false;

/**
 * \brief Default value for the setting "Credit bonus points even if the exam
 * (without bonus points) is assessed as "failed" (5.0)"
 */
inline constexpr bool passDueToBonus = true;

/** \brief Default grade boundary step size  */
inline constexpr GradeBoundaryStepSize boundaryStepSize =
    GradeBoundaryStepSize::Min1_00;

//-----------------------------------------------------------------------

/** \brief Count students with attempt nr 0 as if it were their first attempt
 *
 * Confirmed to be always false after servicedesk-ticket [#20230630-0128].
 */
inline constexpr bool countZerothAttemptAsFirstAttempt = false;

//-----------------------------------------------------------------------

/** \brief Percentage of window width of the grade tab, which is given to the
 * grade histogram. Between 0-100. */
inline constexpr int gradeTabHistogramWidthPercentage = 70;

//-----------------------------------------------------------------------

/** \brief Maximum selectable open points passing limit. */
inline constexpr double openPointsPassingLimit = 0.5;

/** \brief Starting/Default value for the open points passing limit. */
inline constexpr double openPointsStartingPassingLimit = 0.5;

//-----------------------------------------------------------------------

/** \brief Minimum settable point histogram bucket size. */
inline constexpr double minBucketSize = 0.25;
/** \brief Maximum settable point histogram bucket size. */
inline constexpr double maxBucketSize = 5.;
/** \brief Point histogram bucket size selector step size. */
inline constexpr double bucketStep = 0.25;

//-----------------------------------------------------------------------

/** \brief The factor by which the small font size is smaller than the large
 * font size. The small font size is never smaller than the default font size.*/
inline constexpr double smallFontSizeRatio = 0.5;

//-----------------------------------------------------------------------

/** \brief If less than this share of dynexite students are in the RWTHonline or
 * bonus point file, issue a warning */
inline constexpr double studentIntersectWarningThreshold = 0.5;

//-----------------------------------------------------------------------

enum class BoxPlotWhiskers {
    Percentile,
    Iqr,
    MinMax,
};

/** \brief Select with which method box plot whiskers will be calculated. */
inline constexpr BoxPlotWhiskers boxPlotWhiskersMethod =
    BoxPlotWhiskers::Percentile;

/** \brief In the case that the box plot whisker method is "Percentile", choose
 * that percentile (0\<p\<1) for the lower extreme. A sensible value (from the
 * seven-number summary) is 0.0887. */
inline constexpr double boxPlotWhiskerPercentile = 0.0887;

//-----------------------------------------------------------------------

/**
 * \brief Colors used in the app for the grading types open/closed/combined
 */
inline constexpr EnumStorage<QColor> gradingTypeColors = {
    QColor{255, 200, 40}, QColor{0, 150, 255}, QColor{0, 150, 50}};

} // namespace defaults
