#include "data/testable_objects.hpp"

#include "data/graph.hpp"
#include "util/grade_boundary_step.hpp"
#include "util/heuristics.hpp"

#include <gtest/gtest.h>

#include <QList>
#include <QtGlobal>

#include <random>
#include <ranges>

inline constexpr int nrIterations = 100;

using enum GradeBoundaryStepSize;

// Test guessGradeBoundaryStepSize on special ranges, e.g. empty ranges
TEST(HeuristicTest, guessGradeBoundarySpecialRanges) {

    auto res = guessGradeBoundaryStepSize(QList<Points>{});
    EXPECT_EQ(res, Min1_00);

#if RANGE_JOIN_SUPPORTED
    res = guessGradeBoundaryStepSize();
    EXPECT_EQ(res, Min1_00);

    res = guessGradeBoundaryStepSize(QList<Points>{}, QList<Points>{});
    EXPECT_EQ(res, Min1_00);

    res = guessGradeBoundaryStepSize(QList{Points{6.}}, QList<Points>{});
    EXPECT_EQ(res, Min1_00);
#endif
}

// No point value has a fractional part different from 0
TEST(HeuristicTest, guessGradeBoundary1_00) {
    std::default_random_engine e(std::random_device{}());

    std::uniform_int_distribution d(0u, 200u);
    for (int i = 0; i < nrIterations; ++i) {
        QList<Points> v;
        v.resize(d(e));
        for (auto &vecI : v) {
            vecI = Points{static_cast<double>(d(e))};
        }
        const auto res = guessGradeBoundaryStepSize(v);
        EXPECT_EQ(res, Min1_00);
    }
}

// One point value has fractional part 0.50
TEST(HeuristicTest, guessGradeBoundary0_50) {
    std::default_random_engine e(std::random_device{}());
    std::uniform_int_distribution d(1u, 200u);
    for (int i = 0; i < nrIterations; ++i) {
        QList<Points> v;
        v.resize(d(e));
        for (auto &vecI : v) {
            vecI = Points{static_cast<double>(d(e))};
        }
        v.insert(v.begin() + v.size() / 2,
                 Points{static_cast<double>(d(e)) + 0.5});
        const auto res = guessGradeBoundaryStepSize(v);
        EXPECT_EQ(res, Min0_50);
    }
}

// One point value has fractional part 0.25
TEST(HeuristicTest, guessGradeBoundary0_25) {
    std::default_random_engine e(std::random_device{}());
    std::uniform_int_distribution d(1u, 200u);
    for (int i = 0; i < nrIterations; ++i) {
        QList<Points> v;
        v.resize(d(e));
        for (auto &vecI : v) {
            vecI = Points{static_cast<double>(d(e))};
        }
        if (i % 2) {
            v.insert(v.begin() + v.size() / 2,
                     Points{static_cast<double>(d(e)) + 0.25});
        } else {
            v.insert(v.begin() + v.size() / 2,
                     Points{static_cast<double>(d(e)) + 0.75});
        }
        const auto res = guessGradeBoundaryStepSize(v);
        EXPECT_EQ(res, Min0_25);
    }
}

// One point value is not a multiple of 0.25
TEST(HeuristicTest, guessGradeBoundaryOneNonStandard) {
    std::default_random_engine e(std::random_device{}());
    std::uniform_int_distribution d(1u, 200u);
    std::uniform_real_distribution dReal(0., 77.);
    for (int i = 0; i < nrIterations; ++i) {
        QList<Points> v;
        v.resize(d(e));
        for (auto &vecI : v) {
            vecI = Points{static_cast<double>(d(e))};
        }

        // insert a value that is not a multiple of 0.25
        const auto middleIt = v.begin() + v.size() / 2;
        double nonStandardValue = dReal(e);
        if (qFuzzyIsNull(std::fmod(nonStandardValue, 0.25))) {
            nonStandardValue += 0.1;
        }
        v.insert(middleIt, Points{nonStandardValue});

        const auto res = guessGradeBoundaryStepSize(v);
        EXPECT_EQ(res, Min0_25);
    }
}

// All point values are not multiples of 0.25
TEST(HeuristicTest, guessGradeBoundaryAllNonStandard) {
    std::default_random_engine e(std::random_device{}());
    std::uniform_int_distribution d(1u, 200u);
    std::uniform_real_distribution dReal(0., 77.);
    for (int i = 0; i < nrIterations; ++i) {
        QList<Points> v;
        v.resize(d(e));
        for (auto &vecI : v) {
            vecI = Points{dReal(e)};
        }

        // make sure at least one value is not .0 / .5 / .25
        const auto middleIt = v.begin() + v.size() / 2;
        if (qFuzzyIsNull(std::fmod(middleIt->p, 0.25))) {
            v.insert(middleIt, Points{middleIt->p + 0.1});
        }

        const auto res = guessGradeBoundaryStepSize(v);
        EXPECT_EQ(res, Min0_25);
    }
}

// Execute for each DAG
class HeuristicsDagTest
    : public testing::TestWithParam<std::pair<std::string, DAG>> {};

INSTANTIATE_TEST_SUITE_P(HeuristicsDagParameterizedTests, HeuristicsDagTest,
                         testing::ValuesIn(exams()));

// Check the results for our known test files
TEST_P(HeuristicsDagTest, CorrectlyCalculatedInDag) {
    const auto &name = GetParam().first;
    const auto &graph = GetParam().second;

    if (name == "split") {
        const auto res = guessGradeBoundaryStepSize(graph.mExam);
        EXPECT_EQ(res, Min1_00);
    } else if (name == "closed") {
        const auto res = guessGradeBoundaryStepSize(graph.mExam);
        EXPECT_EQ(res, Min0_25);
    } else if (name == "closed_bonus") {
        const auto res = guessGradeBoundaryStepSize(graph.mExam);
        EXPECT_EQ(res, Min0_50);
    } else {
        FAIL() << "Input file with unknown name " << name;
    }
}
